from hashlib import sha256
from time import time

from zope import interface
from zope import component
from zope.i18n import translate
from zope.site.hooks import getSite
from zope.publisher.interfaces import IRequest

from Acquisition import aq_inner
from Products.CMFCore.utils import getToolByName
from Products.Five.browser import BrowserView

from raptus.securelogin import interfaces
from raptus.securelogin import SecureLoginMessageFactory as _

import logging

TOKEN_KEY = 'raptus.securelogin.token'
VALID_KEY = 'raptus.securelogin.valid'


class Secure(object):

    interface.implements(interfaces.ISecure)
    component.adapts(interface.Interface, IRequest)

    def __init__(self, context, request):
        self.request = request
        self.context = getSite()
        self.configuration = interfaces.IConfiguration(getSite())
        self.membership = getToolByName(self.context, 'portal_membership')
        self.registration = getToolByName(self.context, 'portal_registration')
        self.groups = getToolByName(self.context, 'portal_groups')

    def _hash(self, token, seed):
        hash = sha256()
        hash.update(token + seed)
        return hash.digest()

    def enabled(self, username):
        """ Whether secure login is enabled for the provided user
        """
        ip = self.request.get('HTTP_X_FORWARDED_FOR', '')
        if not ip:
            ip = self.request.get('REMOTE_ADDR', '')
        for bypass in self.configuration.ip_bypass:
            if ip.startswith(bypass):
                return False
        for group in self.configuration.groups:
            if username in self.groups.getGroupMembers(group):
                return True
        return False

    def has_token(self, username):
        """ Whether the provided user already received a token
        """
        return int(self.request.SESSION.get(VALID_KEY + '.' + username, 0)) + int(self.configuration.timeout) >= int(time() / 60)

    def check(self, username, token):
        """ Whether the given token is correct for the provided user
        """
        if not self.has_token(username):
            return False
        seed = self.request.SESSION.get(VALID_KEY + '.' + username, '')
        if self._hash(token, seed) == self.request.SESSION.get(TOKEN_KEY + '.' + username, False):
            self.request.SESSION.set(TOKEN_KEY + '.' + username, False)
            self.request.SESSION.set(VALID_KEY + '.' + username, 0)
            return True
        return False

    def send(self, username):
        """ Send a new security token to the provided user
        """
        token = self.registration.getPassword(int(self.configuration.token))
        now = str(int(time() / 60))
        self.request.SESSION.set(TOKEN_KEY + '.' + username, self._hash(token, now))
        self.request.SESSION.set(VALID_KEY + '.' + username, now)
        member = self.membership.getMemberById(username)
        if member is None:
            return
        email = self.configuration.email
        mobile = member.getProperty('mobile', None)
        if email and mobile:
            email = email % dict(number=mobile)
            result = interfaces.SMS
        else:
            email = member.getProperty('email')
            result = interfaces.EMAIL
        mailhost = getToolByName(self.context, 'MailHost')
        if result is interfaces.EMAIL:
            message = translate(_('mail_body',
                                  default='''Use the following token to log in to the site ${site}

    Security token: ${token}

(This token is valid for ${timeout} minutes)''',
                                  mapping={'token': token,
                                           'timeout': self.configuration.timeout,
                                           'site': self.context.getProperty('title')}), context=self.request)
            subject = translate(_('mail_subject',
                                  default='Security token for the site ${site}',
                                  mapping=dict(site=self.context.getProperty('title'))), context=self.request)
        else:
            subject = mobile
            message = token
        email_from_name = self.context.getProperty('email_from_name', '')
        email_from_address = self.context.getProperty('email_from_address', '')
        mailhost.send(message, mto=email, mfrom='%s <%s>' % (email_from_name, email_from_address),
                      subject=subject, charset='utf-8')
        return result



from zope import interface
from zope import component
from zope.formlib import form

from Products.CMFCore.utils import getToolByName
from Products.CMFPlone.utils import safe_unicode
from Products.CMFPlone.interfaces import IPloneSiteRoot
from Products.CMFDefault.formlib.schema import ProxyFieldProperty
from Products.CMFDefault.formlib.schema import SchemaAdapterBase

from plone.app.controlpanel.form import ControlPanelForm

from raptus.securelogin import interfaces
from raptus.securelogin import SecureLoginMessageFactory as _


class SecureLoginControlPanelAdapter(SchemaAdapterBase):

    component.adapts(IPloneSiteRoot)
    interface.implements(interfaces.IConfiguration)

    def __init__(self, context):
        super(SecureLoginControlPanelAdapter, self).__init__(context)
        self.portal = context
        pprop = getToolByName(context, 'portal_properties')
        self.context = pprop.securelogin_properties
        self.encoding = pprop.site_properties.default_charset

    ip_bypass = ProxyFieldProperty(interfaces.IConfiguration['ip_bypass'])
    groups = ProxyFieldProperty(interfaces.IConfiguration['groups'])
    email = ProxyFieldProperty(interfaces.IConfiguration['email'])
    timeout = ProxyFieldProperty(interfaces.IConfiguration['timeout'])
    token = ProxyFieldProperty(interfaces.IConfiguration['token'])


class SecureLoginControlPanel(ControlPanelForm):

    form_fields = form.FormFields(interfaces.IConfiguration)

    label = _(u'Secure login settings')
    description = ''
    form_name = ''



from zope import component

from Acquisition import aq_base
from AccessControl.SecurityInfo import ClassSecurityInfo
from Globals import InitializeClass

from Products.PlonePAS.plugins.cookie_handler import ExtendedCookieAuthHelper
from Products.PluggableAuthService.interfaces.plugins import IValidationPlugin
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from Products.statusmessages.interfaces import IStatusMessage

from raptus.securelogin import interfaces
from raptus.securelogin import SecureLoginMessageFactory as _


def manage_addSecureLoginCookieAuthHelper(self, id, title='',
                                          RESPONSE=None, **kw):
    """Create an instance of a secure login cookie auth helper.
    """
    self = self.this()

    o = SecureLoginCookieAuthHelper(id, title, **kw)
    self._setObject(o.getId(), o)
    o = getattr(aq_base(self), id)

    if RESPONSE is not None:
        RESPONSE.redirect('manage_workspace?manage_tabs_message=Secure+Login+Authentication+added.')


manage_addSecureLoginCookieAuthHelperForm = PageTemplateFile("zmi/SecureLoginCookieAuthHelperForm", globals())


class SecureLoginCookieAuthHelper(ExtendedCookieAuthHelper):
    meta_type = "Secure Login Authentication"

    security = ClassSecurityInfo()

    security.declarePrivate('extractCredentials')
    def extractCredentials(self, request):
        """ Extract credentials from cookie or 'request'. """
        login = request.form.get('__ac_name', '')
        if login:
            secure = component.getMultiAdapter((self, request), interfaces.ISecure)
            if not secure.enabled(login):
                return ExtendedCookieAuthHelper.extractCredentials(self, request)
            token = request.form.get('__ac_securitytoken', '')
            if not request.SESSION.get('secureloginenabled', False) or not secure.has_token(login):
                if secure.send(login) is interfaces.EMAIL:
                    msg = _('message_email',
                            default=u'A security token has been sent to your email address, '
                                     'please enter it into the provided field to complete '
                                     'the login.')
                else:
                    msg = _('message_mobile',
                            default=u'A security token has been sent to your mobile number, '
                                     'please enter it into the provided field to complete '
                                     'the login.')
                IStatusMessage(request).add(msg, 'info')
                request.form['secureloginfailure'] = True
                request.SESSION.set('secureloginenabled', True)
                return {}
            if secure.check(login, token):
                request.SESSION.set('secureloginenabled', False)
                return ExtendedCookieAuthHelper.extractCredentials(self, request)
            else:
                request.form['secureloginfailure'] = True
                IStatusMessage(request).add(_('error_token', default=u'The provided security token is not valid.'), 'error')
            return {}
        return ExtendedCookieAuthHelper.extractCredentials(self, request)


InitializeClass(SecureLoginCookieAuthHelper)



from Products.CMFCore.utils import getToolByName
from Products.PlonePAS.Extensions.Install import setupAuthPlugins, activatePluginInterfaces

from raptus.securelogin import auth

PROPERTIES = (
    ('ip_bypass', 'lines', ''),
    ('groups', 'lines', ''),
    ('email', 'string', ''),
    ('timeout', 'int', 5),
    ('token', 'int', 8)
)

PLUGIN_ID = 'secureloginauth'


def setupVarious(context):
    if context.readDataFile('raptus.securelogin_various.txt') is None:
        return

    portal = context.getSite()

    properties = getToolByName(portal, 'portal_properties').securelogin_properties
    for id, type, default in PROPERTIES:
        if not properties.hasProperty(id):
            properties._setProperty(id, default, type)

    acl = getToolByName(portal, 'acl_users')

    if acl.objectIds(['Extended Cookie Auth Helper']):
        disable=['IExtractionPlugin', 'ICredentialsResetPlugin', 'ICredentialsUpdatePlugin']
        activatePluginInterfaces(portal, 'credentials_cookie_auth',
                disable=disable)
    if not acl.objectIds(['Secure Login Authentication']):
        auth.manage_addSecureLoginCookieAuthHelper(acl, 'securelogin_credentials_cookie_auth', cookie_name='__ac')
        disable=['ICredentialsResetPlugin', 'ICredentialsUpdatePlugin', 'IChallengePlugin']
        activatePluginInterfaces(portal, 'securelogin_credentials_cookie_auth',
                disable=disable)


def uninstallVarious(context):
    if context.readDataFile('raptus.securelogin_uninstall.txt') is None:
        return

    portal = context.getSite()

    cp = getToolByName(portal, 'portal_controlpanel')
    if 'SecureLoginSettings' in cp:
        cp.unregisterConfiglet('SecureLoginSettings')

    acl = getToolByName(portal, 'acl_users')

    if acl.objectIds(['Secure Login Authentication']):
        acl.manage_delObjects(['securelogin_credentials_cookie_auth'])

    disable=['ICredentialsResetPlugin', 'ICredentialsUpdatePlugin']
    activatePluginInterfaces(portal, 'credentials_cookie_auth',
            disable=disable)



from zope import interface
from zope import schema

from plone.app.users.userdataschema import IUserDataSchemaProvider, IUserDataSchema
from plone.app.users.browser.personalpreferences import UserDataPanelAdapter

from raptus.securelogin import SecureLoginMessageFactory as _


class MobileUserDataSchemaProvider(object):
    interface.implements(IUserDataSchemaProvider)

    def getSchema(self):
        """
        """
        return IMobileUserDataSchema


class IMobileUserDataSchema(IUserDataSchema):
    """ Add a field to enter the mobile phone number used to send the
        security token
    """

    mobile = schema.TextLine(
        title=_(u'Mobile'),
        description=_('description_mobile', default=u'Your mobile phone number able to receive SMS messages.'),
        required=False
    )


class MobileUserDataPanelAdapter(UserDataPanelAdapter):
    interface.implements(IMobileUserDataSchema)

    def get_mobile(self):
        return self._getProperty('mobile')

    def set_mobile(self, value):
        if value is None:
            value = ''
        return self.context.setMemberProperties({'mobile': value.replace(' ', '').replace('+', '00')})

    mobile = property(get_mobile, set_mobile)



from zope import component

from raptus.securelogin import interfaces


def get_secure(context, request):
    return component.getMultiAdapter((context, request), interfaces.ISecure)


def enabled(context, request):
    return get_secure(context, request).enabled()


def check(context, request, token):
    return get_secure(context, request).check(token)


def send(context, request):
    return get_secure(context, request).send()



from zope.i18nmessageid import MessageFactory

from Products.CMFCore.permissions import ManagePortal
from Products.PluggableAuthService import registerMultiPlugin

SecureLoginMessageFactory = MessageFactory('raptus.securelogin')

from raptus.securelogin import auth


#################################
# register plugins with pas
try:
    registerMultiPlugin(auth.SecureLoginCookieAuthHelper.meta_type)
except RuntimeError:
    # make refresh users happy
    pass


def initialize(context):

    context.registerClass(auth.SecureLoginCookieAuthHelper,
                          permission=ManagePortal,
                          constructors=(auth.manage_addSecureLoginCookieAuthHelperForm,
                                        auth.manage_addSecureLoginCookieAuthHelper),
                          visibility=None
                          )



# -*- coding: utf-8 -*-
from zope import interface
from zope import schema

from raptus.securelogin import SecureLoginMessageFactory as _

EMAIL = object()
SMS = object()


class InvalidEmailAddress(schema.ValidationError):
    __doc__ = _('error_pattern', default=u'Email pattern has to contain «%(number)s»')


def validateEmail(value):
    if not '%(number)s' in value:
        raise InvalidEmailAddress(value)
    return True


class IConfiguration(interface.Interface):

    ip_bypass = schema.List(
        title=_(u'IP'),
        description=_('description_ip_bypass', default=u'IPs for which to bypass the secure login procedure.'),
        required=False,
        value_type=schema.TextLine()
    )

    groups = schema.Set(
        title=_(u'Groups'),
        description=_('description_groups', default=u'Groups for which to enforce the secure login procedure.'),
        value_type=schema.Choice(
            source='plone.app.vocabularies.Groups',
        ),
        required=False
    )

    email = schema.TextLine(
        title=_(u'Email pattern'),
        description=_('description_email', default=u'The email address (has to contain «%(number)s», which will '
                      'be replaced by the mobile phone number of the user) to send '
                      'the security token to. If omitted the security token will be '
                      'sent to the email address of the user.'),
        required=False,
        constraint=validateEmail
    )

    timeout = schema.Int(
        title=_(u'Validity'),
        description=_('description_timeout', default=u'Number of minutes a security token is valid.'),
        default=5,
        required=True
    )

    token = schema.Int(
        title=_(u'Token length'),
        description=_('description_token', default=u'Length of the generated security token.'),
        default=8,
        required=True
    )


class ISecure(interface.Interface):

    def enabled(username):
        """ Whether secure login is enabled for the provided user
        """

    def has_token(username):
        """ Whether the provided user already received a token
        """

    def check(username, token):
        """ Whether the given token is correct for the provided user
        """

    def send(username):
        """ Send a new security token to the provided user
        """
